import { BaseService } from '@/application/base/baseService'
import config from '@/application/config'
import { SocketEvent } from '@/application/enums'
import type { MapEventTileWithTeleport, UUID } from '@/application/types'
import { Character } from '@/entities/character'
import MapManager from '@/managers/mapManager'
import SocketManager from '@/managers/socketManager'
import MapEventTileRepository from '@/repositories/mapEventTileRepository'
import TeleportService from '@/services/characterTeleportService'
import type { Server } from 'socket.io'

type Position = { positionX: number; positionY: number }
export type Node = Position & { parent?: Node; g: number; h: number; f: number }

class CharacterMoveService extends BaseService {
  private io: Server = SocketManager.getIO()

  // Rotation lookup table for better performance
  private readonly ROTATION_MAP = {
    diagonal: {
      'up-left': 7,
      'down-right': 3,
      'up-right': 5,
      'down-left': 1
    },
    straight: {
      left: 6,
      right: 2,
      down: 4,
      up: 0
    }
  }

  private readonly DIRECTIONS = [
    { x: 0, y: -1 }, // Up
    { x: 0, y: 1 }, // Down
    { x: -1, y: 0 }, // Left
    { x: 1, y: 0 }, // Right
    { x: -1, y: -1 }, // Up left
    { x: -1, y: 1 }, // Up right
    { x: 1, y: -1 }, // Down left
    { x: 1, y: 1 } // Down right
  ]

  public async calculatePath(character: Character, targetX: number, targetY: number): Promise<Position[] | null> {
    if (!this.validateInput(character, targetX, targetY)) {
      return null
    }

    const map = MapManager.getMapById(character.map.id)
    const grid = await map?.getGrid()

    if (!grid?.length) {
      return null
    }

    // Validate target position is within map boundaries
    if (targetX < 0 || targetY < 0 || targetX >= grid[0]!.length || targetY >= grid.length) {
      return null
    }

    // Check if target position is walkable
    if (grid[Math.floor(targetY)]![Math.floor(targetX)] !== 0) {
      return null
    }

    // Ensure we're working with valid coordinates
    const start: Position = {
      positionX: Math.floor(character.positionX),
      positionY: Math.floor(character.positionY)
    }

    const end: Position = {
      positionX: Math.floor(targetX),
      positionY: Math.floor(targetY)
    }

    // Don't calculate path if start and end are the same
    if (start.positionX === end.positionX && start.positionY === end.positionY) {
      return null
    }

    return this.findPath(start, end, grid)
  }

  private validateInput(character: Character, targetX: number, targetY: number): boolean {
    if (!character || !character.map) {
      this.logger.error('Invalid character or map data')
      return false
    }

    if (!Number.isFinite(targetX) || !Number.isFinite(targetY)) {
      this.logger.error('Invalid target coordinates')
      return false
    }

    return true
  }

  public calculateRotation(X1: number, Y1: number, X2: number, Y2: number): number {
    if (config.ALLOW_DIAGONAL_MOVEMENT) {
      // Check diagonal movements using lookup table
      if (X1 > X2 && Y1 > Y2) return this.ROTATION_MAP.diagonal['up-left']
      if (X1 < X2 && Y1 < Y2) return this.ROTATION_MAP.diagonal['down-right']
      if (X1 > X2 && Y1 < Y2) return this.ROTATION_MAP.diagonal['up-right']
      if (X1 < X2 && Y1 > Y2) return this.ROTATION_MAP.diagonal['down-left']
    }

    // Non-diagonal movements using lookup table
    if (X1 > X2) return this.ROTATION_MAP.straight.left
    if (X1 < X2) return this.ROTATION_MAP.straight.right
    if (Y1 < Y2) return this.ROTATION_MAP.straight.down
    if (Y1 > Y2) return this.ROTATION_MAP.straight.up

    return this.ROTATION_MAP.straight.up // Default case
  }

  private findPath(start: Position, end: Position, grid: number[][]): Node[] {
    const openList = new PriorityQueue<Node>((a, b) => a.f - b.f)
    const closedSet = new Set<string>()
    const MAX_ITERATIONS = 1000
    let iterations = 0

    const getKey = (p: Position) => `${p.positionX},${p.positionY}`
    const gScoreMap = new Map<string, number>()

    openList.enqueue({ ...start, g: 0, h: this.getDistance(start, end), f: this.getDistance(start, end) })
    gScoreMap.set(getKey(start), 0)

    try {
      while (openList.length > 0 && iterations < MAX_ITERATIONS) {
        iterations++
        const current = openList.dequeue()
        if (!current) break

        if (current.positionX === end.positionX && current.positionY === end.positionY) {
          return this.reconstructPath(current)
        }

        const currentKey = getKey(current)
        closedSet.add(currentKey)

        const neighbors = this.getValidNeighbors(current, grid, end)

        for (const neighbor of neighbors) {
          const neighborKey = getKey(neighbor)
          if (closedSet.has(neighborKey)) continue

          const tentativeGScore = current.g + this.getDistance(current, neighbor)

          if (!gScoreMap.has(neighborKey) || tentativeGScore < gScoreMap.get(neighborKey)!) {
            gScoreMap.set(neighborKey, tentativeGScore)
            const h = this.getDistance(neighbor, end)
            const f = tentativeGScore + h

            const node: Node = { ...neighbor, g: tentativeGScore, h, f, parent: current }
            openList.enqueue(node)
          }
        }
      }

      this.logger.warn(`Path not found after ${iterations} iterations`)
      return []
    } finally {
      openList.clear()
      closedSet.clear()
      gScoreMap.clear()
    }
  }

  private getValidNeighbors(current: Position, grid: number[][], end: Position): Position[] {
    return this.DIRECTIONS.slice(0, config.ALLOW_DIAGONAL_MOVEMENT ? 8 : 4)
      .map((dir) => ({
        positionX: current.positionX + dir.x,
        positionY: current.positionY + dir.y
      }))
      .filter((pos) => this.isValidPosition(pos, grid, end))
  }

  private isValidPosition(pos: Position, grid: number[][], end: Position): boolean {
    if (pos.positionX < 0 || pos.positionY < 0 || pos.positionX >= grid[0]!.length || pos.positionY >= grid.length) {
      return false
    }

    return grid[pos.positionY]![pos.positionX] === 0 || (pos.positionX === end.positionX && pos.positionY === end.positionY)
  }

  isValidStep(current: { positionX: number; positionY: number }, next: { positionX: number; positionY: number }): boolean {
    return Math.abs(next.positionX - current.positionX) <= 1 && Math.abs(next.positionY - current.positionY) <= 1
  }

  async checkMapEvents(map_id: UUID, nextTile: { positionX: number; positionY: number }) {
    const mapEventTileRepository = new MapEventTileRepository()
    return mapEventTileRepository.getEventTileByMapIdAndPosition(map_id, nextTile.positionX, nextTile.positionY)
  }

  async handleTeleportMapEventTile(character_id: UUID, mapEventTile: MapEventTileWithTeleport): Promise<void> {
    const teleport = mapEventTile.getTeleport()
    if (teleport) {
      await TeleportService.teleportCharacter(character_id, {
        targetMapId: teleport.getToMap().getId(),
        targetX: teleport.getToPositionX(),
        targetY: teleport.getToPositionY(),
        rotation: teleport.getToRotation()
      })
    }
  }

  private getDistance(a: Position, b: Position): number {
    const dx = Math.abs(a.positionX - b.positionX)
    const dy = Math.abs(a.positionY - b.positionY)
    // Manhattan distance for straight paths, then Euclidean for diagonals
    return dx + dy + (Math.sqrt(2) - 2) * Math.min(dx, dy)
  }

  private reconstructPath(endNode: Node): Node[] {
    const path: Node[] = []
    let current: Node | undefined = endNode
    while (current) {
      path.unshift(current)
      current = current.parent
    }
    return path
  }

  public broadcastMovement(character: Character, isMoving: boolean): void {
    this.io.in(character.map.id).emit(SocketEvent.MAP_CHARACTER_MOVE, [character.id, character.getPositionX(), character.getPositionY(), character.getRotation(), isMoving])
  }

  public validateMovementDistance(
    currentX: number,
    currentY: number,
    lastKnownPosition: { x: number; y: number } | null,
    stepDelay: number,
    isMoving: boolean
  ): { isValid: boolean; maxAllowedDistance: number; actualDistance: number } {
    if (!lastKnownPosition || !isMoving) {
      return { isValid: true, maxAllowedDistance: 0, actualDistance: 0 }
    }

    const maxAllowedDistance = Math.ceil((stepDelay / 1000) * (config.ALLOW_DIAGONAL_MOVEMENT ? 2.5 : 2))
    const actualDistance = Math.sqrt(Math.pow(currentX - lastKnownPosition.x, 2) + Math.pow(currentY - lastKnownPosition.y, 2))

    return {
      isValid: actualDistance <= maxAllowedDistance,
      maxAllowedDistance,
      actualDistance
    }
  }
}

export default new CharacterMoveService()

class PriorityQueue<T> {
  private items: T[] = []
  constructor(private compare: (a: T, b: T) => number) {}

  enqueue(item: T): void {
    this.items.push(item)
    this.siftUp(this.items.length - 1)
  }

  dequeue(): T | undefined {
    if (this.items.length === 0) return undefined

    const result = this.items[0]
    const last = this.items.pop()

    if (this.items.length > 0 && last !== undefined) {
      this.items[0] = last
      this.siftDown(0)
    }

    return result
  }

  get length(): number {
    return this.items.length
  }

  private siftUp(index: number): void {
    while (index > 0) {
      const parentIndex = Math.floor((index - 1) / 2)
      if (this.compare(this.items[index]!, this.items[parentIndex]!) < 0) {
        ;[this.items[index], this.items[parentIndex]] = [this.items[parentIndex]!, this.items[index]!]
        index = parentIndex
      } else {
        break
      }
    }
  }

  private siftDown(index: number): void {
    const length = this.items.length
    while (true) {
      let minIndex = index
      const leftChild = 2 * index + 1
      const rightChild = 2 * index + 2

      if (leftChild < length && this.compare(this.items[leftChild]!, this.items[minIndex]!) < 0) {
        minIndex = leftChild
      }
      if (rightChild < length && this.compare(this.items[rightChild]!, this.items[minIndex]!) < 0) {
        minIndex = rightChild
      }

      if (minIndex === index) break
      ;[this.items[index], this.items[minIndex]] = [this.items[minIndex]!, this.items[index]!]
      index = minIndex
    }
  }

  clear(): void {
    this.items = []
  }
}
